#ifndef wt_table_H
#define wt_table_H

/* Weight tables for filtering digital signals */
/* Last edited on 2009-02-23 22:07:50 by stolfi */

#define wt_table_H_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#define _GNU_SOURCE
#include <stdio.h>

#include <vec.h>
#include <bool.h>
#include <argparser.h>

/* WEIGHT TABLES WITHOUT ALLOCATION

  The procedures in this section create filter weight tables of 
  arbitrary length {n>0}, even or odd, stored into client-given arrays.  */

void wt_table_fill_gaussian(double sigma, int n, double wt[]);
  /* Stores into {wt[0..n-1]} a table derived from a Gaussian
    distribution {G(z)} with mean {n/2} and standard deviation {sigma}.
    
    More precisely, {wt[i]} will be the integral of {G(z)} over 
    the interval {[i _ i+1]}. The table is normalized so that
    the sum of all entries is 1. */

void wt_table_fill_binomial(int n, double wt[]);
  /* Sets {wt[i] = choose(i,n-1)/2^(n-1)} for {i} in {0..n-1}.
    The variance will be {(n-1)/4}, so the
    standard deviation will be {sqrt(n-1)/2)}. */

void wt_table_fill_triangular(int n, double wt[]);
  /* Builds a triangular weight table, {w[i] = (r+1 - |i-r|)/(r+1)^2}
    for {i} in {0..n-1}, where {r = (n-1)/2}. Requires {n} odd. This
    is the convolution of two uniform distributions on {r+1}
    elements. */

/* WEIGHT TABLES WITH ALLOCATION

  The procedures in this section create weight tables of 
  odd length {n = 2*r+1}, packaged as newly allocated {double_vec_t}s. 
  
  !!! These procedure should be generalized to even lengths. !!! */

double_vec_t wt_table_make_gaussian(double sigma, double maxLoss);
  /* Builds a weight table derived from a Gaussian distribution {G(z)}
    of standard deviation {sigma}. The table length will be odd,
    {2*r+1} for some {r>=0}. The procedure chooses {r} so that the
    total omitted (out-of-window) probability does not exceed
    {maxLoss}.
    
    More precisely, {w[i]} will be the integral of {G(z)} within a
    unit interval centered on {i-r}, for {i = 0..2*r}. The omitted
    probability is the integral of {G(z)} outside the interval
    {[-(r+1/2) _ +(r+1/2)]}.
    
    The table is normalized so that the sum of all entries is 1. */

double_vec_t wt_table_make_binomial(int r);
  /* Builds a weight table derived from a binomial distribution {w[i]
    = choose(i,n)/2^n} for {i} in {0..n}, where {n = 2*r}. Note that
    this is the convolution of {2*r} binary distributions. The table
    length will be {2*r+1}. The variance will be {r/2}, so the
    standard deviation will be {sqrt(r/2)}. */

double_vec_t wt_table_make_triangular(int r);
  /* Builds a triangular weight table, {w[i] = (r+1 - |i-r|)/(r+1)^2}
    for {i} in {0..2*r}. Note that this is the convolution of two
    uniformdistributions on {r+1} elements. The table length will be
    {2*r+1}. */

/* PRINTOUT */

void wt_table_print(FILE *wr, char *wtname, int n, double wt[]);
  /* prints to {wr} the name {wtname} (if not NULL), the weight table
    {wt[0..n-1]}, and its main statistical properties. */

char *wt_table_make_descr(int n, double wt[], char *fmt);
  /* Returns the weight table {wt[0..n-1]} formatted as a character string,
    with each element printed with the given {fmt}. */

/* PARSING COMMAND LINE ARGUMENTS */

double_vec_t wt_table_args_parse(argparser_t *pp, bool_t unitNorm);
   /* Parses a sequence of numbers from the command line, in the
     format described by {wt_table_args_HELP} below. \
     
     If {unitNorm} is TRUE, scales all weights so that their sum is 1;
     the optional "/ {DENOM}" is parsed but ignored. 
     
     If {unitNorm} is FALSE, does no normalization; but, if the
     optional "/ {DENOM}" is present, divides all weights by
     {DENOM}. */
     
#define wt_table_args_HELP \
  "{WEIGHT}.. [ \"/\" {DENOM} ]"

#define wt_table_args_INFO \
  "The weights may be integer or fractional," \
  " and may include \"e\" exponent parts. If the" \
  " optional slash and denominator are present," \
  " all weights are divided by {DENOM}."

#define wt_table_args_norm_sum_INFO \
  "The weights may be integer or fractional," \
  " and may include \"e\" exponent parts. They are" \
  " implicitly normalized to unit sum. (The slash" \
  " and {DENOM}, if present, are parsed but ignored.)"

#endif
